#include <src/donnees/fonctions/fonctionreelle.h>
#include <src/donnees/instructions/instructiondo.h>
#include <src/donnees/instructions/instructionreturn.h>
#include <src/donnees/fonctions/instancenormale.h>
#include <src/donnees/fonctions/instancerecursive.h>
#include <src/donnees/fonctions/declarations.h>
#include <src/donnees/types/typevoid.h>
#include <src/donnees/expressions/expressionimage.h>
#include <src/donnees/expressions/expressionnoyau.h>

#include <algorithm>
using namespace Traceur::Donnees::Fonction;

FonctionReelle::FonctionReelle(const PrototypeSP& prototype, const Traceur::Donnees::Instruction::InstructionSP& corps, const Traceur::Donnees::Source::PositionSP& position)
	:Fonction(prototype,position), _type(Type::NORMALE)
{
	Traceur::Donnees::Instruction::DoSP instrdo = std::dynamic_pointer_cast<Traceur::Donnees::Instruction::Do>(corps);
	if(instrdo)
	{
		_corps = instrdo->_bloc;
	}
	else
	{
		Traceur::Donnees::Instruction::BlocSP bloc(new Traceur::Donnees::Instruction::Bloc());
		bloc->ajoute(corps);
		bloc->mode_restreint();
		_corps = bloc;
	}
	
	if(Traceur::Donnees::Type::Void::SP()->compatible(prototype->_type_retour))
	{
		Traceur::Donnees::Instruction::ReturnSP ret(new Traceur::Donnees::Instruction::Return(position));
		_corps->ajoute(ret);
	}

	if(not _corps->teste_retour_fonction(prototype->_type_retour))
	{
		throw Traceur::Exception::Retour_manquant(position,prototype);
	}
}

bool FonctionReelle::recursive() const
{
	if(_corps and _corps->_declarations)
		return _corps->_declarations->recursive();
	return _type==Type::RECURSIVE;
}

void FonctionReelle::disp(std::ostream& os) const
{
	_appel.lock()->disp(os);
	os << std::string(*_position);
	format_disp(os);
}

void FonctionReelle::format_disp(std::ostream& os) const
{
	os << "func ";
	if(recursive())
		os << "rec ";
	_prototype->format(os);
	os << " := " << std::endl << "do" << std::endl;
	_corps->format(os);
	os << "end" << std::endl;
}

void FonctionReelle::format(std::ostream& os) const
{
#ifdef TRACEUR_DEBUG
	os << "Real function: ";
	if(recursive())
		os << "rec ";
	_prototype->format(os);
	os << std::endl;
#endif
	_corps->format(os);
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction
		{
			struct CreationLienParametre
			{
				CreationLienParametre(PrototypeSP prototype, FonctionReelle::LiensParametres& parametres,Traceur::Donnees::Instruction::BlocSP corps, Traceur::Donnees::Source::PositionSP position)
					:_prototype(prototype),_parametres(parametres),_corps(corps),_position(position) {};
				void operator() (const PrototypeSP prototype)
				{
					InstanceSP instance(new InstanceNormale());
					_parametres.push_back(instance);
					if(_corps->_declarations->ajoute_declaration(prototype,instance))
						throw Traceur::Exception::ParametreAmbigu(_position,prototype,_prototype);
				};

				~CreationLienParametre() {};
				private:
				PrototypeSP _prototype;
				FonctionReelle::LiensParametres& _parametres;
				Traceur::Donnees::Instruction::BlocSP _corps;
				Traceur::Donnees::Source::PositionSP _position;
			};
		}
	}
}

void FonctionReelle::edition_liens(Traceur::Donnees::Instruction::BlocSP bloc)
{
	_corps->cree_declarations(bloc);

	if(_type==Type::RECURSIVE)
	{
		InstanceSP instance(new InstanceRecursive(this->shared_from_this()));
		_corps->_declarations->ajoute_declaration(_prototype,instance);
	}

	std::for_each(_prototype->_parametres_gauche.cbegin(),_prototype->_parametres_gauche.cend(),CreationLienParametre(_prototype,_parametres_gauche,_corps,_position));
	std::for_each(_prototype->_parametres_droite.cbegin(),_prototype->_parametres_droite.cend(),CreationLienParametre(_prototype,_parametres_droite,_corps,_position));

	_corps->edition_liens(bloc);
	_corps->_fonction = this->shared_from_this();
	_type=Type::NORMALE;
}

void FonctionReelle::definition_fonction()
{
	_corps->definition_fonction();
	_corps->simplification();
}

namespace Traceur
{
	namespace Donnees
	{
		namespace Fonction
		{
			struct SauvegardeParametres
			{
				FonctionSP operator() (InstanceSP instance)
				{
					return instance->valeur_dynamique();
				}
			};
		}
	}
}

void FonctionReelle::empile_parametres(Traceur::Donnees::Fonction::Fonction::ListeParametres gauche, Traceur::Donnees::Fonction::Fonction::ListeParametres droite, PileParametres& sauvegarde_gauche, PileParametres& sauvegarde_droite)
{
	std::transform(_parametres_gauche.cbegin(),_parametres_gauche.cend(),std::back_inserter(sauvegarde_gauche),SauvegardeParametres());
	std::transform(_parametres_droite.cbegin(),_parametres_droite.cend(),std::back_inserter(sauvegarde_droite),SauvegardeParametres());

	{// valeur_dynamique est une methode virtuelle, donc on peut pas appliquer transform ici, apparemment
		auto itliens=_parametres_gauche.begin();
		for(auto iter=gauche.begin(); iter!=gauche.end() ; ++iter, ++itliens)
		{
			(*itliens)->valeur_dynamique(*iter);
		}
	}
	{
		auto itliens=_parametres_droite.begin();
		for(auto iter=droite.begin(); iter!=droite.end() ; ++iter, ++itliens)
		{
			(*itliens)->valeur_dynamique(*iter);
#ifdef TRACEUR_DEBUG
			TRACEUR_TRACE << (*iter) << std::endl;
#endif
		}
	}
}

void FonctionReelle::depile_parametres(PileParametres& sauvegarde_gauche, PileParametres& sauvegarde_droite)
{
	{
		auto itliens = _parametres_gauche.begin();
		for(auto iter = sauvegarde_gauche.begin() ; iter!=sauvegarde_gauche.end() ; ++iter,++itliens)
		{
			(*itliens)->valeur_dynamique(*iter);
		}
	}
	{
		auto itliens = _parametres_droite.begin();
		for(auto iter = sauvegarde_droite.begin() ; iter!=sauvegarde_droite.end() ; ++iter,++itliens)
		{
			(*itliens)->valeur_dynamique(*iter);
		}
	}
}
					
Traceur::Donnees::Instruction::InstructionInterruptionSP FonctionReelle::appel(
	ListeParametres gauche, ListeParametres droite, Traceur::Donnees::Assembleur::AssembleurSP& assembleur,
						Traceur::Donnees::Expression::AppelFonctionUtilisateurSP appel)
{
	Traceur::Donnees::Global::ClonesInstances clone_instances;
	FonctionSP clone = this->clone(clone_instances);
	FonctionReelleSP valeur_appel = std::dynamic_pointer_cast<FonctionReelle>(clone);
	if(not valeur_appel) throw;

	valeur_appel->_appel = appel;

	{
		auto itliens=_parametres_gauche.begin();
		for(auto iter=gauche.begin(); iter!=gauche.end() ; ++iter, ++itliens)
		{
			clone_instances[(*itliens).get()]->valeur_dynamique(*iter);
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "appel param gauche> " << clone_instances[(*itliens).get()];
#endif
		}
	}
	{
		auto itliens=_parametres_droite.begin();
		for(auto iter=droite.begin(); iter!=droite.end() ; ++iter, ++itliens)
		{
			clone_instances[(*itliens).get()]->valeur_dynamique(*iter);
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "appel param droite> " << clone_instances[(*itliens).get()];
#endif
		}
	}

#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << "valeur appel> " << valeur_appel;
#endif
	
	return valeur_appel->_corps->evaluation(assembleur);
}

Traceur::Donnees::Expression::ExpressionSP FonctionReelle::substitution(ListeParametres gauche, ListeParametres droite)
{
	if(_type==Type::RECURSIVE)
		return Traceur::Donnees::Expression::ExpressionSP();
	
	Traceur::Donnees::Expression::ExpressionSP expression = _corps->commence_par_return();
	if(not expression)
		return Traceur::Donnees::Expression::ExpressionSP();
	
	PileParametres sauvegarde_gauche;
	PileParametres sauvegarde_droite;

	empile_parametres(gauche,droite,sauvegarde_gauche,sauvegarde_droite);

	Traceur::Donnees::Global::ClonesInstances clones_instances;
	Traceur::Donnees::Expression::ExpressionSP clone = expression->clone(clones_instances);

	Traceur::Donnees::Expression::ExpressionSP retour = clone->simplification();
	
#ifdef TRACEUR_DEBUG
	TRACEUR_TRACE << _prototype << " subst " << retour << std::endl;
#endif

	depile_parametres(sauvegarde_gauche,sauvegarde_droite);

	return retour;
}

FonctionSP FonctionReelle::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	FonctionSP clone_ptetre = clones_instances[this];
	if(clone_ptetre)
		return clone_ptetre;
	PrototypeSP prototype = _prototype->clone(clones_instances);
	FonctionReelleSP clone(new FonctionReelle(_type,prototype,_position));
	clones_instances.ajout(this,clone);
	Traceur::Donnees::Instruction::BlocSP corps = _corps->clone(clones_instances);
	clone->_corps = corps;
	std::transform(_parametres_gauche.begin(),_parametres_gauche.end(),std::back_inserter(clone->_parametres_gauche),bind(&Instance::clone,std::placeholders::_1,std::ref(clones_instances)));
	std::transform(_parametres_droite.begin(),_parametres_droite.end(),std::back_inserter(clone->_parametres_droite),bind(&Instance::clone,std::placeholders::_1,std::ref(clones_instances)));
	clone->_appel = _appel;
	return clone;
}
