#include <src/donnees/instructions/bloc.h>

#include <src/donnees/assembleur/assembleur.h>
#include <src/donnees/fonctions/declarationsrestreintes.h>
#include <src/donnees/fonctions/declarationsisolees.h>
#include <src/donnees/instructions/valeur.h>
#include <src/donnees/instructions/instructionreturn.h>
#include <src/donnees/expressions/expression.h>

#include <algorithm>

using namespace Traceur::Donnees::Instruction;

void Bloc::ajoute(InstructionSP instruction)
{
	_instructions.push_back(instruction);
}

bool Bloc::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	return std::find_if(_instructions.cbegin(),_instructions.cend(),bind(&Instruction::teste_retour_fonction,std::placeholders::_1,std::ref(type))) != _instructions.cend();
}

void Bloc::disp(std::ostream& os) const
{
	Traceur::Donnees::Fonction::FonctionSP fonction = _fonction.lock();
	if(fonction)
	{
#ifdef TRACEUR_DEBUG
		std::cerr << "BLOC-FONCTION\n";
#endif
		fonction->disp(os);
		os << std::endl;
	}
	else
	{
		InstructionSP instruction = _parent.lock();
		if(instruction)
		{
#ifdef TRACEUR_DEBUG
			std::cerr << "BLOC-INSTRUCTION\n";
#endif
			instruction->disp(os);
			os << std::endl;
		}
#ifdef TRACEUR_DEBUG
		else
		{
			std::cerr << "BLOC-RIENDUTOUT\n";
		}
#endif
	}
	if(_declarations)
	{
		_declarations->disp(os);
	}
}

void Bloc::format(std::ostream& os) const
{
#ifdef TRACEUR_DEBUG
	os << "Bloc " ;
	switch(_type)
	{
		case Type::NORMAL: os << "normal" ; break ;
		case Type::RESTREINT: os << "restricted" ; break ;
		case Type::ISOLE: os << "isolated" ; break ;
	}
	os << std::endl;
	if(_declarations)
		_declarations->format(os);
#endif
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::format,std::placeholders::_1,std::ref(os)));
#ifdef TRACEUR_DEBUG
	os << "End of bloc" << std::endl;
#endif
}

void Bloc::edition_liens(BlocSP parent, InstructionSP instruction)
{
	if(instruction)
		_parent = instruction;
	
	cree_declarations(parent);

	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::edition_liens,std::placeholders::_1,this->shared_from_this()));
}

Traceur::Donnees::Expression::ExpressionSP Bloc::commence_par_return()
{
	if(_instructions.empty())
		return Traceur::Donnees::Expression::ExpressionSP();
	ReturnSP retour = std::dynamic_pointer_cast<Return>(*(_instructions.begin()));
	if(not retour)
		return Traceur::Donnees::Expression::ExpressionSP();
	return retour->_valeur;
}

void Bloc::simplification()
{
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::simplification,std::placeholders::_1));
}

InstructionInterruptionSP Bloc::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	assembleur->debut_bloc(_declarations);
	for(auto instriter = _instructions.begin() ; instriter != _instructions.end() ; ++instriter)
	{
#ifdef TRACEUR_DEBUG
		if((*instriter)->position())
		{
			std::string str = *((*instriter)->position());
			TRACEUR_TRACE << "Instruction position: " <<  str << std::endl;
		}
		else
		{
			TRACEUR_TRACE << "Instruction position: none" << std::endl;
		}
#endif
		InstructionInterruptionSP interruption = (*instriter)->evaluation(assembleur);
		if(interruption)
		{
			assembleur->bloc_interrompu();
			return interruption;
		}
	}
	assembleur->fin_bloc();
	return InstructionInterruptionSP();
}

void Bloc::cree_declarations(BlocSP parent)
{
	if(_declarations)
		return;

	Traceur::Donnees::Fonction::DeclarationsSP declarations;

	switch(_type)
	{
		case Type::NORMAL:
			if(parent)
				declarations = Traceur::Donnees::Fonction::DeclarationsSP(new Traceur::Donnees::Fonction::Declarations(parent->_declarations));
			else
				declarations = Traceur::Donnees::Fonction::DeclarationsSP(new Traceur::Donnees::Fonction::Declarations(Traceur::Donnees::Fonction::DeclarationsSP()));
			break;
		case Type::RESTREINT:
			if(parent)
				declarations = Traceur::Donnees::Fonction::DeclarationsSP(new Traceur::Donnees::Fonction::DeclarationsRestreintes(parent->_declarations));
			else
				declarations = Traceur::Donnees::Fonction::DeclarationsSP(new Traceur::Donnees::Fonction::DeclarationsRestreintes(Traceur::Donnees::Fonction::DeclarationsSP()));
			break;
			break;
		case Type::ISOLE:
			if(parent)
				declarations = Traceur::Donnees::Fonction::DeclarationsSP(new Traceur::Donnees::Fonction::DeclarationsIsolees(parent->_declarations));
			else
				declarations = Traceur::Donnees::Fonction::DeclarationsSP(new Traceur::Donnees::Fonction::DeclarationsIsolees(Traceur::Donnees::Fonction::DeclarationsSP()));
			break;
			break;
	}
	
	_declarations = declarations;
}

void Bloc::definition_fonction()
{
	std::for_each(_instructions.begin(),_instructions.end(),bind(&Instruction::definition_fonction,std::placeholders::_1));
}

BlocSP Bloc::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	InstructionSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent = _parent.lock();
	Traceur::Donnees::Fonction::FonctionSP fonction = clones_instances[_fonction.lock().get()];
	if(not fonction)
		fonction = _fonction.lock();
	BlocSP clone(new Bloc());
	clone->_parent = parent;
	clone->_fonction = fonction;
	clone->_type=_type;
	clones_instances.ajout(this,clone);
	if(_declarations) // les fonctions expr n'ont pas de declarations
		clone->_declarations = _declarations->clone(clones_instances);
	std::transform(_instructions.begin(),_instructions.end(),std::back_inserter(clone->_instructions),bind(&Instruction::clone,std::placeholders::_1,std::ref(clones_instances)));
	return clone;
}

#ifdef TRACEUR_DEBUG
void debug_bloc(Traceur::Donnees::Instruction::BlocSP bloc)
{
	bloc->format(std::cout);
}
#endif
