#include <src/donnees/instructions/instructionif.h>
#include <src/donnees/instructions/instructionerror.h>
#include <src/donnees/types/typeboolean.h>
#include <src/donnees/expressions/expressionbooleenvrai.h>
#include <src/donnees/expressions/expressionbooleenpeutetre.h>
#include <algorithm>

using namespace Traceur::Donnees::Instruction;

struct TestConditions
{
	bool operator() (If::ClauseVraie clause)
	{
		return not (clause.first->est_de_type<Traceur::Donnees::Type::Boolean>());
	}
};

	If::If(ClausesVraies clauses_vraies, const BlocSP& clause_fausse,const Traceur::Donnees::Source::PositionSP& position)
:InstructionControle(position),_clauses_vraies(clauses_vraies),_clause_fausse(clause_fausse)
{
	auto condition_non_booleene = std::find_if(clauses_vraies.begin(), clauses_vraies.end(),TestConditions());
	if(condition_non_booleene!=clauses_vraies.end())
	{
		throw Traceur::Exception::ConditionNonBooleenne(_position,condition_non_booleene->first->type());
	}
}

struct TestRetourFonction
{
	TestRetourFonction(const Traceur::Donnees::Type::TypeSP& type)
		:_type(type) {};
	bool operator() (If::ClauseVraie clause)
	{
		return not clause.second->teste_retour_fonction(_type);
	}
	private:
		Traceur::Donnees::Type::TypeSP _type;
};

bool If::teste_retour_fonction(const Traceur::Donnees::Type::TypeSP& type) const
{
	auto it = std::find_if(_clauses_vraies.cbegin(),_clauses_vraies.cend(),TestRetourFonction(type));
	return (it==_clauses_vraies.cend()) and (_clause_fausse->teste_retour_fonction(type));
}

void If::disp(std::ostream& os) const
{
	Instruction::disp(os);
	format(os);
}

void If::format(std::ostream& os) const
{
	os << "if ";
	_clauses_vraies[0].first->format(os);
	os << " then" << std::endl;
	_clauses_vraies[0].second->format(os);
	auto it = _clauses_vraies.cbegin();
	for(++it ; it!=_clauses_vraies.cend() ; ++it)
	{
		os << "elif ";
		(it->first)->format(os);
		os << " then" << std::endl;
		(it->second)->format(os);
	}
	os << "else" << std::endl;
	_clause_fausse->format(os);
	os << "end" << std::endl;
}

struct EditionLiens
{
	EditionLiens(InstructionSP instruction, BlocSP bloc)
		:_instruction(instruction),_bloc(bloc) {};
	void operator() (If::ClauseVraie clause_vraie)
	{
		clause_vraie.first->edition_liens(_instruction);
		clause_vraie.second->edition_liens(_bloc,_instruction);
	};
	private:
		InstructionSP _instruction;
		BlocSP _bloc;
};

void If::edition_liens(BlocSP bloc)
{
	Instruction::edition_liens(bloc);
	std::for_each(_clauses_vraies.begin(),_clauses_vraies.end(),EditionLiens(this->shared_from_this(),bloc));
	_clause_fausse->edition_liens(bloc,this->shared_from_this());
}


InstructionInterruptionSP If::evaluation(Traceur::Donnees::Assembleur::AssembleurSP& assembleur)
{
	for(auto iter=_clauses_vraies.begin() ; iter!=_clauses_vraies.end() ; ++iter)
	{
		ValeurSP condition = iter->first->evaluation(assembleur);
		if(not condition->a_une_valeur_retour())
		{
			InstructionInterruptionSP interruption = *condition;
			interruption->ajoute_position(_position);
			return interruption;
		}
		Traceur::Donnees::Expression::ExpressionSP expr_condition = *condition;
		if(expr_condition->est_de_nature<Traceur::Donnees::Expression::BooleenVrai>())
		{
			InstructionInterruptionSP vraie = iter->second->evaluation(assembleur);
			if(vraie)
				vraie->ajoute_position(_position);
			return vraie;
		}
		if(expr_condition->est_de_nature<Traceur::Donnees::Expression::BooleenPeutetre>())
		{
			return Error::condition_peut_etre_trouvee(_position);
		}
	}
	InstructionInterruptionSP fausse=_clause_fausse->evaluation(assembleur);
	if(fausse)
		fausse->ajoute_position(_position);
	return fausse;
}

struct DefinitionFonction
{
	void operator() (If::ClauseVraie clause_vraie)
	{
		clause_vraie.first->definition_fonction();
		clause_vraie.second->definition_fonction();
	};
};

void If::definition_fonction()
{
	std::for_each(_clauses_vraies.begin(),_clauses_vraies.end(),DefinitionFonction());
	_clause_fausse->definition_fonction();
}

struct CloneClauseVraie
{
	CloneClauseVraie(Traceur::Donnees::Global::ClonesInstances& clones_instances)
	:_clones_instances(clones_instances) {};
	If::ClauseVraie operator() (If::ClauseVraie clause_vraie)
	{
		If::ClauseVraie clone;
		clone.first = clause_vraie.first->clone(_clones_instances);
		clone.second = clause_vraie.second->clone(_clones_instances);
		return clone;
	};
	private:
		Traceur::Donnees::Global::ClonesInstances& _clones_instances;
};

InstructionSP If::clone(Traceur::Donnees::Global::ClonesInstances& clones_instances)
{
	BlocSP parent = clones_instances[_parent.lock().get()];
	if(not parent)
		parent=_parent.lock();
	IfSP clone(new If(_position));
	clone->_parent=parent;
	clones_instances.ajout(this,clone);
	std::transform(_clauses_vraies.begin(),_clauses_vraies.end(),std::back_inserter(clone->_clauses_vraies),CloneClauseVraie(clones_instances));
	clone->_clause_fausse = _clause_fausse->clone(clones_instances);
	clone->_parent = parent;
	return clone;
}
